/**
  * @file
  *
  * @author Nadi Sarrar <nadi@net.t-labs.tu-berlin.de>
  *
  * @section LICENSE
  *
  * libTFO, a flow offloading decision engine designed for switch controllers.
  *
  * Copyright (C) 2010, 2011  Nadi Sarrar <nadi@net.t-labs.tu-berlin.de>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * @section DESCRIPTION
  *
  * A simple application for testing libTFO. Uses string as type for matches
  * and actions, reads input from stdin.
  */

#include <libtfo/tfo.hh>
#include <libtfo/extras.hh>

#include <iostream>
#include <iomanip>

using namespace std;

static void fib_entry_dump_head (void)
{
	cout
		<< setw(11) << "match "
		<< setw(11) << "action "
		<< setw(11) << "state "
		<< setw(7) << "table "
		<< setw(4) << "ttl "
		<< setw(11) << "total "
		<< setw(9) << "b1 "
		<< setw(9) << "b2 "
		<< setw(8) << "b3"
		<< endl;
}

static void fib_entry_dump (void*, const FibEntry<string, string, int>* e, void*)
{
	cout
		<< setw(10) << e->match() << " "
		<< setw(10) << e->action() << " "
		<< setw(10) << fib_entry_state_str(e->state()) << " "
		<< setw(6) << fib_entry_table_str(e->table()) << " "
		<< setw(3) << e->ttl() << " "
		<< setw(10) << e->hits_total() << " "
		<< setw(8) << e->hits_b1() << " "
		<< setw(8) << e->hits_b2() << " "
		<< setw(8) << e->hits_b3()
		<< endl;
}

static void fib_modify (void*, FibEntry<string, string, int>* e)
{
	cout << "[@" << fib_entry_table_str(e->table()) << ": "
		<< fib_entry_state_str(e->state())
		<< " " << e->match() << " => " << e->action()
		<< "]" << endl;
}

struct hashmatchp
{
	inline size_t operator()(const string* s1) const
	{
		hash<char*> H;
		return H(s1->c_str());
	}
};

struct eqmatchp
{
	inline bool operator()(const string* s1, const string* s2) const
	{
		return (*s1 == *s2);
	}
};

typedef TFO<void*, string, string, int, hashmatchp, eqmatchp> MyTFO;

#define BOLD "\x1b[1m"
#define NORMAL "\x1b[0m"

int main (int argc, char* argv[])
{
	int fib_capacity = argc > 1 ? atoi(argv[1]) : 10;
	int size_b2 = argc > 2 ? atoi(argv[2]) : 2;
	int size_b3 = argc > 3 ? atoi(argv[3]) : 4;

	cout << "*** SETTINGS: ";
	cout << "fib_capacity=" << BOLD << fib_capacity << NORMAL;
	cout << " size_b2=" << BOLD << size_b2 << NORMAL;
	cout << " size_b3=" << BOLD << size_b3 << NORMAL;
	cout << endl;

	MyTFO* fv = new MyTFO(NULL, fib_capacity, fib_modify, size_b2, size_b3);

	string command;
	string match;
	string data;

	int time = 0;

	for (;;)
	{
		int new_time;
		cin >> new_time;

		if (!time)
		{
			time = new_time;
		}
		else if (new_time != time)
		{
			cout << "--------------------------------- ";
			cout << BOLD << "FIB at time " << new_time << NORMAL;
			cout << " ---------------------------------" << endl;
			fv->bin_complete();
			fib_entry_dump_head();
			fv->walk_entries(fib_entry_dump);

			if (++time != new_time)
			{
				cout << "Invalid input.." << endl;
				break;
			}
		}

		cin >> command >> match >> data;
		if (!cin.good())
			break;

		switch (command[0])
		{
		case 'A':
			fv->update_entry(match, data, FIB_ENTRY_ADD);
			break;
		case 'D':
			fv->update_entry(match, data, FIB_ENTRY_REMOVE);
			break;
		case 'M':
			fv->update_entry(match, data, FIB_ENTRY_MODIFY);
			break;
		case 'S':
			fv->update_hits(match, atoi(data.c_str()));
			break;
		}
	}

	delete fv;

	return 0;
}
