/**
  * @file
  *
  * @author Nadi Sarrar <nadi@net.t-labs.tu-berlin.de>
  *
  * @section LICENSE
  *
  * libTFO, a flow offloading decision engine designed for switch controllers.
  *
  * Copyright (C) 2010, 2011  Nadi Sarrar <nadi@net.t-labs.tu-berlin.de>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * @section DESCRIPTION
  *
  * FibEntry, the type of forwarding table entries used by libTFO.
  */

#ifndef __FIBENTRY_HH__
#define __FIBENTRY_HH__

#include <stdint.h>

enum FibEntryState
{
	FIB_ENTRY_STATE_INACTIVE, /**< Entry removed, still keeping the statistics. */
	FIB_ENTRY_STATE_ACTIVE, /**< Entry is active and used in forwarding. */
	FIB_ENTRY_STATE_MOD_ADD, /**< Active entry added. */
	FIB_ENTRY_STATE_MOD_MOVE, /**< Active entry switched tables. */
	FIB_ENTRY_STATE_MOD_ACTION, /**< Active entry changed action. */
	FIB_ENTRY_STATE_MOD_REMOVE, /**< Active entry removed. */
};

enum FibEntryTable
{
	FIB_ENTRY_TABLE_FAST,
	FIB_ENTRY_TABLE_SLOW
};

enum FibEntryHitsUpdateType
{
	FIB_ENTRY_HITS_ADD,
	FIB_ENTRY_HITS_UPDATE_BIN,
	FIB_ENTRY_HITS_UPDATE_TOTAL,
	FIB_ENTRY_HITS_REPLACE
};

/**
  * @brief Forwarding table entry along with its traffic statistics.
  *
  * @tparam M Type of flow match, e.g., destination IP prefix.
  * @tparam A Type of flow action, e.g., output interface.
  * @tparam U Type of additional information attached to FIB entries.
  */
template <typename M, typename A, typename U>
class FibEntry
{
	public:
		typedef M match_type;
		typedef A action_type;
		typedef U user_type;

		/**
		  * Creates a new instance of FibEntry.
		  *
		  * @param[in] size_b2 Number of b1 time bins that make a b2 time bin.
		  * @param[in] size_b3 Number of b1 time bins that make a b3 time bin.
		  * @param[in] m Match.
		  * @param[in] a Action.
		  */
		inline FibEntry (uint32_t size_b2, uint32_t size_b3, const M& m, const A& a)
			: _match(m), _action(a), _old_action(a),
			_state(FIB_ENTRY_STATE_INACTIVE), _table(FIB_ENTRY_TABLE_SLOW), _ttl(0),
			_size_b2(size_b2), _size_b3(size_b3),
			_hits_current(0), _hits_total(0), _hits_b1(0), _hits_b2(0), _hits_b3(0),
			_hits_ring_size(0), _hits_ring_pos(0)
		{
			_hits_ring = new uint64_t[size_b3];
		}

		/**
		  * Deletes a FibEntry.
		  */
		inline ~FibEntry (void)
		{
			delete [] _hits_ring;
		}

		/**
		  * @return Returns the match.
		  */
		inline const M& match (void) const
		{ return _match; }

		/**
		  * @return Returns the action.
		  */
		inline const A& action (void) const
		{ return _action; }

		/**
		  * Sets the action.
		  *
		  * @param[in] a Action.
		  *
		  * @return Returns the action.
		  */
		inline const A& action (const A& a)
		{ return (_action = a); }

		/**
		  * @return Returns the action prior to the last modification.
		  */
		inline const A& old_action (void) const
		{ return _old_action; }

		/**
		  * Sets the action prior to the last modification.
		  *
		  * @param[in] a Action.
		  *
		  * @return Returns the action prior to the last modification.
		  */
		inline const A& old_action (const A& a)
		{ return (_old_action = a); }

		/**
		  * @return User data accessor.
		  */
		inline U& user_data (void)
		{ return _user_data; }

		/**
		  * @return Returns the FibEntryState.
		  */
		inline const enum FibEntryState& state (void) const
		{ return _state; }

		/**
		  * Sets the FibEntryState.
		  *
		  * @param[in] s State.
		  *
		  * @return Returns the FibEntryState.
		  */
		inline const enum FibEntryState& state (enum FibEntryState s)
		{ return (_state = s); }

		/**
		  * @return Returns the FibEntryTable.
		  */
		inline const enum FibEntryTable& table (void) const
		{ return _table; }

		/**
		  * Sets the FibEntryTable.
		  *
		  * @param[in] t Table.
		  *
		  * @return Returns the FibEntryTable.
		  */
		inline const enum FibEntryTable& table (enum FibEntryTable t)
		{ return (_table = t); }

		/**
		  * @return Returns the TTL value.
		  */
		inline const uint32_t ttl (void) const
		{ return _ttl; }

		/**
		  * Sets the TTL.
		  *
		  * @param[in] t TTL value.
		  *
		  * @return Returns the TTL value.
		  */
		inline const uint32_t ttl (uint32_t t)
		{ return (_ttl = t); }

		/**
		  * Decrements the TTL.
		  *
		  * @return Returns the TTL value.
		  */
		inline const uint32_t ttl_decrement (void)
		{ return (_ttl ? --_ttl : 0); }

		/**
		  * @return Returns the total number of hits.
		  */
		inline const uint64_t& hits_total (void) const
		{ return _hits_total; }

		/**
		  * @return Returns the hits of the most recently completed b1 time bin.
		  */
		inline const uint64_t& hits_b1 (void) const
		{ return _hits_b1; }

		/**
		  * @return Returns the hits of the most recently completed b2 time bin.
		  */
		inline const uint64_t& hits_b2 (void) const
		{ return _hits_b2; }

		/**
		  * @return Returns the hits of the most recently completed b3 time bin.
		  */
		inline const uint64_t& hits_b3 (void) const
		{ return _hits_b3; }

		/**
		  * Modifies the hit counter. Call bin_complete() when b1 has finished.
		  *
		  * @param[in] hits Number of hits.
		  * @param[in] type Type of update.
		  */
		inline void update_hits (const uint64_t& hits, const enum FibEntryHitsUpdateType type)
		{
			switch (type)
			{
			case FIB_ENTRY_HITS_ADD:
				_hits_current += hits;
				break;
			case FIB_ENTRY_HITS_UPDATE_BIN:
				_hits_current = hits;
				break;
			case FIB_ENTRY_HITS_UPDATE_TOTAL:
				if (hits >= _hits_total)
				{
					_hits_current = hits - _hits_total;
					break;
				}
				/* if (hits < _hits_total), we assume a counter wrap-around or reset
				   has happened, i.e., we proceed similar to FIB_ENTRY_HITS_REPLACE. */
			case FIB_ENTRY_HITS_REPLACE:
				_hits_current = hits;
				_hits_total = 0;
				break;
			}
		}

		/**
		  * Updates the hit counters of all time scales (b1, b2, b3).
		  */
		inline void bin_complete (void)
		{
		/** @cond */
#define FIB_ENTRY_RING_POS(curr, dist, size) \
			(((curr) >= (dist)) ? \
			 ((curr) - (dist)) : \
			 ((size) - ((dist) - (curr))))
		/** @endcond */

			_hits_total += _hits_current;

			_hits_b1 = _hits_current;

			_hits_b2 += _hits_current;
			if (_hits_ring_size >= _size_b2)
				_hits_b2 -= _hits_ring[FIB_ENTRY_RING_POS(_hits_ring_pos, _size_b2 - 1, _hits_ring_size)];

			_hits_b3 += _hits_current;
			if (_hits_ring_size >= _size_b3)
				_hits_b3 -= _hits_ring[FIB_ENTRY_RING_POS(_hits_ring_pos, _size_b3 - 1, _hits_ring_size)];

			_hits_ring_pos = (_hits_ring_pos + 1) % _size_b3;
			_hits_ring[_hits_ring_pos] = _hits_current;
			if (_hits_ring_size < _size_b3)
				_hits_ring_size++;

			_hits_current = 0;
		}

	private:
		/* forwarding information */
		const M _match;
		A _action;
		A _old_action;
		U _user_data;

		/* installation status */
		enum FibEntryState _state;
		enum FibEntryTable _table;
		int _ttl;

		/* bin sizes */
		uint32_t _size_b2;
		uint32_t _size_b3;

		/* traffic statistics */
		uint64_t  _hits_current;
		uint64_t  _hits_total;
		uint64_t  _hits_b1;
		uint64_t  _hits_b2;
		uint64_t  _hits_b3;
		uint64_t* _hits_ring;
		uint32_t  _hits_ring_size;
		uint32_t  _hits_ring_pos;
};

#endif
